"""Exceptions ViewSets"""

from typing import Any
from rest_framework.status import HTTP_400_BAD_REQUEST
from six import text_type


class JsonException(Exception):
    """Custom exception for JSON responses."""

    def __init__(self, message: Any, status_code: int = HTTP_400_BAD_REQUEST) -> None:
        self.status_code = status_code
        self.message = (
            {"errors": text_type(message)}
            if not isinstance(message, (dict, list))
            else message
        )
        super().__init__(self.message)



"""Importer"""

from typing import Any
from importlib import import_module


def import_module_attribute(import_string: str, package: str = None) -> Any:
    """Import a module or attribute from a module."""
    try:
        return import_module(import_string, package)
    except ImportError:
        module, attr = import_string.rsplit(".", 1)
        return getattr(import_module(module, package), attr)



# -*- coding: utf-8 -*-
from RestrictedPython import compile_restricted, safe_globals
from swifty.expressionator.namespace import (
    ExpressionatorNamespace,
    ExpressionatorException,
    ALLOWED_BUILTINS,
    ALLOWED_NAMES,
)
from swifty.logging.logger import SwiftyLoggerMixin
from swifty.utils.objects import DictObject


class ExpressionatorManager(SwiftyLoggerMixin):
    """_summary_

    Raises:
        error: _description_
        ExpressionatorException: _description_

    Returns:
        _type_: _description_
    """

    expressionator_namespace = None
    data = None
    __expr_data = None

    def __init__(self, data=None, namespace=ExpressionatorNamespace) -> None:
        """_summary_

        Args:
            data (_type_, optional): _description_. Defaults to None.
        """
        self.expr_data = data or {}
        self.safe_expr_namespace = namespace().safe_expr_namespace or {}
        self.safe_locals = {
            **ALLOWED_NAMES,
            **ALLOWED_BUILTINS,
            **self.expr_data,
            **self.safe_expr_namespace,
        }

    @property
    def expr_data(self):
        """_summary_

        Returns:
            _type_: _description_
        """
        return self.__expr_data

    @expr_data.setter
    def expr_data(self, data):
        if isinstance(data, DictObject):
            self.__expr_data = data
        elif isinstance(data, dict):
            self.__expr_data = DictObject(dict_data=data)
        else:
            raise ExpressionatorException("Invalid formatted data")

    def __expressionator(self, expression, default=None, strict=False):
        """
        _summary_

        Args:
            expression (_type_): _description_

        Raises:
            error: _description_

        Returns:
            _type_: _description_
        """
        try:
            default = self.evaluate(expression)
        except (ExpressionatorException, NameError) as error:
            self.logger.error(
                "An error occurred before evaluating the expression",
                expression=expression,
                default=default,
                error=error,
            )
        except Exception as error:
            self.logger.error(
                "An error occurred when evaluating the expression",
                expression=expression,
                default=default,
                error=error,
            )
            if strict:
                raise error
        return default

    def get_value(self, field_name):
        """
        _summary_

        Args:
            field_name (_type_): _description_

        Returns:
            _type_: _description_
        """
        return self.__expressionator(expression=field_name)

    def initialator(
        self,
        initialization_data,
    ):
        """
        _summary_

        Args:
            initialization_data (_type_): _description_

        Returns:
            _type_: _description_
        """
        expression = initialization_data.get("expression")
        default = initialization_data.get("default")
        return self.__expressionator(expression=expression, default=default)

    def calculator(self, calculation_data, data=None):
        """_summary_

        Args:
            calculation_data (_type_): _description_
            data (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_
        """
        if data is not None:
            self.expr_data = data
            self.safe_locals.update(self.expr_data)
        expression = calculation_data.get("expression")
        default = calculation_data.get("default")
        return self.__expressionator(expression=expression, default=default)

    def validator(self, validation_data, value):
        """_summary_

        Args:
            validation_data (_type_): _description_
            value (_type_): _description_

        Returns:
            _type_: _description_
        """
        self.safe_locals.update({"value": value})
        expression = validation_data.get("expression")
        return self.__expressionator(expression=expression, default=False)

    def evaluate(self, expression):
        """
        Evaluate a math expression.

        Args:
            expression (_type_): _description_

        Raises:
            ExpressionatorException: _description_

        Returns:
            _type_: _description_
        """
        # Compile the expression
        code = compile_restricted(expression, "<string>", "eval")

        # Validate allowed names
        for name in code.co_names:
            if name not in self.safe_locals:
                raise ExpressionatorException(f"The use of '{name}' is not allowed")

        return eval(code, safe_globals, self.safe_locals)



"""ViewSets"""

from typing import Any, List
from six import text_type
from django.contrib.auth.models import User
from django.http import JsonResponse, HttpRequest
from rest_framework import status
from rest_framework.viewsets import ViewSet
from rest_framework.renderers import JSONRenderer
from rest_framework.response import Response
from rest_framework.exceptions import NotAuthenticated, AuthenticationFailed
from rest_framework.permissions import BasePermission
from rest_framework.authentication import BaseAuthentication
from rest_framework_simplejwt.authentication import JWTAuthentication
from rest_framework_simplejwt.exceptions import InvalidToken, TokenError

from swifty.viewsets.constants import LogEvents
from swifty.viewsets.exceptions import JsonException
from swifty.contexts.request_context import RequestContext
from swifty.logging.logger import SwiftyLoggerMixin
from swifty.auth.permissions import SwiftyPermission


class SwiftyViewSet(ViewSet, SwiftyLoggerMixin):
    """A view set that integrates JWT authentication and custom logging."""

    response: Response
    renderer_classes: tuple = (JSONRenderer,)
    authentication_classes: tuple = (JWTAuthentication,)
    permission_classes: tuple = (SwiftyPermission,)

    def initialize_request(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpRequest:
        """Initializes the request object.

        Args:
            request (HttpRequest): The incoming request.

        Returns:
            HttpRequest: The initialized request.
        """
        if not getattr(request, "is_initialized", False):
            request = super().initialize_request(request, *args, **kwargs)
            setattr(request, "is_initialized", True)
        return request

    def dispatch(self, request: HttpRequest, *args: Any, **kwargs: Any) -> Response:
        """Dispatches the request to the appropriate handler.

        Args:
            request (HttpRequest): The incoming request.

        Returns:
            Response: The response from the view.
        """
        request = self.initialize_request(request, *args, **kwargs)
        with RequestContext(logger=self.logger, request_data=request):
            return super().dispatch(request, *args, **kwargs)

    def handle_exception(self, exc: Exception) -> Response:
        """Handles exceptions raised during request processing.

        Args:
            exc (Exception): The exception raised.

        Returns:
            Response: A JSON response with the error details.
        """
        self.logger.error(LogEvents.EXCEPTION_HANDLER, exceptions=text_type(exc))
        if isinstance(exc, JsonException):
            return JsonResponse(
                getattr(exc, "message", exc.__dict__),
                status=getattr(
                    exc, "status_code", status.HTTP_500_INTERNAL_SERVER_ERROR
                ),
            )
        return super().handle_exception(exc)

    def get_permissions(self) -> List[BasePermission]:
        """Returns the list of permissions required for this view.

        Returns:
            List[BasePermission]: The list of permission instances.
        """
        action = self.action and getattr(self, self.action, None)
        permission_classes = (
            getattr(action, "permission_classes", self.permission_classes)
            if action
            else ()
        )
        return [permission() for permission in permission_classes or ()]

    def get_authenticators(self) -> List[BaseAuthentication]:
        """Returns the list of authenticators that this view can use.

        Returns:
            List[BaseAuthentication]: The list of authenticator instances.
        """
        action = getattr(
            self, self.action_map.get(self.request.method.lower(), ""), None
        )  # type: ignore
        authentication_classes = (
            getattr(action, "authentication_classes", self.authentication_classes)
            if action
            else ()
        )
        return [auth() for auth in authentication_classes or ()]

    def perform_authentication(self, request: HttpRequest) -> User:
        """Performs authentication on the request.

        Args:
            request (HttpRequest): The incoming request.

        Raises:
            NotAuthenticated: If authentication fails.

        Returns:
            User: The authenticated user.
        """
        if request.authenticators:
            try:
                getattr(request, "_authenticate")()
            except (TokenError, InvalidToken) as exc:
                raise AuthenticationFailed from exc

            if not request.user or not request.user.is_authenticated:
                raise NotAuthenticated

            return request.user



"""
This module provides a monkey patch for the inspect module to replace
the deprecated getargspec function with a compatible alternative using
signature.
"""

import inspect
from collections import namedtuple

# Save the original getargspec function
original_getargspec = inspect.getargspec

# Define a custom named tuple to mimic the FullArgSpec structure
FullArgSpec = namedtuple(
    "FullArgSpec",
    [
        "args",
        "varargs",
        "varkw",
        "defaults",
        "kwonlyargs",
        "kwonlydefaults",
        "annotations",
    ],
)


# Create a replacement function for getargspec
def patched_getargspec(func):
    """Return a FullArgSpec for the given function."""
    signature = inspect.signature(func)

    # Extract the positional arguments and keyword-only arguments
    args = [
        param.name
        for param in signature.parameters.values()
        if param.default == inspect.Parameter.empty
    ]
    kwonlyargs = [
        param.name
        for param in signature.parameters.values()
        if param.default != inspect.Parameter.empty
        and param.kind == inspect.Parameter.KEYWORD_ONLY
    ]

    # Build the FullArgSpec object with the necessary information
    return FullArgSpec(
        args=args,
        varargs=None,
        varkw=None,
        defaults=None,
        kwonlyargs=kwonlyargs,
        kwonlydefaults=None,
        annotations={
            param.name: param.annotation
            for param in signature.parameters.values()
            if param.annotation is not inspect.Parameter.empty
        },
    )


# Apply the monkey patch
inspect.getargspec = patched_getargspec
print("Monkey patch applied: getargspec is now using signature.")



"""Cache Manager"""

from django.conf import settings

from django_auxilium.utils.functools import (
    BaseCache,
    Caching,
    MemoizeDecorator,
    MemoizeDescriptor,
    NotInCache,
)

from swifty.logging.logger import SwiftyLoggerMixin
from swifty.utils.mapper import getpath

from swifty.utils.importer import import_module_attribute
from swifty.cache.constants import SwiftyCacheEvents


class SwiftyCacheDescriptor(MemoizeDescriptor, SwiftyLoggerMixin):
    """Descriptor for caching with custom parameters."""

    cache_attribute_pattern: str = "{name}"

    def __init__(
        self,
        *args,
        ttl: int = None,
        static_key: str = None,
        vary_by: tuple = (),
        context_getter: str = None,
        **kwargs,
    ):
        self.ttl = ttl

        self.static_key = static_key

        self.vary_by = vary_by
        self.context_getter = context_getter
        super().__init__(*args, **kwargs)

    @staticmethod
    def default_cache_class(*args, **kwargs):
        """Instantiate cache backend."""
        return get_cache_backend()(*args, **kwargs)  # type: ignore

    def get_cache(self, instance) -> Caching:
        """Get cache implementation with custom parameters."""

        return self.cache_class(
            instance,
            self.cache_attribute,
            ttl=self.ttl or getpath(instance, "redis_cache_pattern.ttl", 3600),
            static_key=self.static_key,  # type: ignore
            vary_by=(
                self.vary_by or getpath(instance, "redis_cache_pattern.vary", "")
            ),  # type: ignore
            context_getter=self.context_getter,  # type: ignore
        )

    def getter(self, instance, *args, new_cache: bool = None, **kwargs) -> any:
        """Return cached value or compute and store it."""

        cache_instance = self.get_cache(instance)

        try:
            if new_cache:

                self.logger.debug(
                    f"{SwiftyCacheEvents.CREATE_NEW_CACHE}",
                    cached_function=getpath(cache_instance, "attr", None),
                    args=args,
                    kwargs=kwargs,
                )
                raise NotInCache

            cached_data = cache_instance.get(*args, **kwargs)
            self.logger.debug(
                f"{SwiftyCacheEvents.CACHED_DATA}",
                cached_function=getpath(cache_instance, "attr", None),
                data=cached_data,
                args=args,
                kwargs=kwargs,
            )
            return cached_data

        except NotInCache:

            self.logger.debug(
                f"{SwiftyCacheEvents.DATA_NOT_IN_CACHE}",
                caching_function=getpath(cache_instance, "attr", None),
                args=args,
                kwargs=kwargs,
            )

            return cache_instance.set(
                self.method(instance, *args, **kwargs), *args, **kwargs
            )


class SwiftyCacheDecorator(MemoizeDecorator, SwiftyLoggerMixin):
    """Decorator for caching with custom parameters."""

    cache_descriptor_class = SwiftyCacheDescriptor

    def __init__(
        self,
        ttl: int = None,
        static_key: str = None,
        vary_by: tuple = (),
        context_getter: str = "context",
        is_method: bool = False,
        as_property: bool = False,
    ):
        self.ttl = ttl
        self.static_key = static_key

        self.vary_by = vary_by

        self.context_getter = context_getter

        self.as_property = as_property

        super().__init__(is_method)

    @staticmethod
    def cache_class(*args, **kwargs):
        """Instantiate cache backend."""
        return get_cache_backend()(*args, **kwargs)  # type: ignore

    def get_cache_descriptor(self) -> SwiftyCacheDescriptor:
        """Instantiate cache descriptor with custom parameters."""

        return self.cache_descriptor_class(
            self.to_wrap,
            ttl=self.ttl,
            static_key=self.static_key,
            vary_by=self.vary_by,
            context_getter=self.context_getter,
            as_property=self.as_property,
        )


def get_cache_backend() -> BaseCache:
    """Get the correct cache backend as configured in settings."""

    if not hasattr(settings, "CACHE_BACKEND"):
        # pylint: disable=import-outside-toplevel
        from swifty.redis.cache import RedisCache

        return RedisCache

    return import_module_attribute(settings.CACHE_BACKEND)


cache = SwiftyCacheDecorator.as_decorator()
cache_property = SwiftyCacheDecorator.as_decorator(as_property=True)



"""Constants"""

import json
from typing import Dict, Tuple, Any
from datetime import date
from rest_framework import serializers
from rest_framework.fields import empty
from rest_framework.exceptions import ValidationError
from django.core.exceptions import ValidationError as DjangoValidationError
from swifty.expressionator.manager import ExpressionatorManager
from swifty.utils.helpers import validation_error_detail, json_handler
from swifty.serializers.constants import EMPTY_VALUES
from swifty.sqldb.core import Base


class BasicFieldSerializer:
    """
    _summary_

    Args:
        object (_type_): _description_
    """

    static_args = (
        "read_only",
        "write_only",
        "required",
        "default",
        "initial",
        "source",
        "label",
        "help_text",
        "style",
        "error_messages",
        "validators",
        "allow_null",
    )
    extra_args: Tuple[str, ...] = tuple()
    default: Any = None
    data = None
    default_empty_html = None

    def __init__(self, *args, extra_kwargs=None, required=None, **kwargs):
        self.extra_kwargs = extra_kwargs
        # Set the default value based on the presence of 'required'
        default = (
            kwargs.pop("default", self.default)
            if required
            else kwargs.get("default", self.default)
        )

        # Initialize the parent class with required
        super().__init__(*args, required=required, **kwargs)

        # Update the default if it was empty
        if self.default is empty:
            self.default = default

    def bind(self, field_name, parent):
        """_summary_

        Args:
            field_name (_type_): _description_
            parent (_type_): _description_
        """
        super().bind(field_name, parent)
        self.initiate_kwargs()

    @property
    def expressionator(self) -> ExpressionatorManager:
        """_summary_

        Args:
            data (_type_): _description_

        Returns:
            _type_: _description_
        """

        expressionator = (
            ExpressionatorManager
            if isinstance(self.root, BasicFieldSerializer)
            else getattr(self.root, "expressionator", ExpressionatorManager)
        )
        if callable(expressionator):
            return expressionator(data=getattr(self.root, "initial_data", {}))
        return expressionator

    def initiate_kwargs(self):
        """_summary_

        Args:
            data (_type_): _description_
        """
        initialator = self.extra_kwargs.get("initialator")
        if initialator:
            initial_attrs = (
                self.expressionator.initialator(initialization_data=initialator) or {}
            )
            field_args = self.static_args + self.extra_args
            for attr, value in initial_attrs.items():
                if attr in field_args:
                    setattr(self, attr, value)
                else:
                    self.extra_kwargs[attr] = value

    def is_disabled_by_rules(self, enabled_rules, value=empty):
        """_summary_

        Args:
            enabled_rules (_type_): _description_

        Returns:
            _type_: _description_
        """
        is_disabled = (
            self.extra_kwargs.get("disabled")
            or (enabled_rules and value in [None, empty])
            or False
        )
        if not is_disabled and isinstance(enabled_rules, (list, tuple)):
            for rule in enabled_rules:
                valid_values = rule.get("values")
                field_value = self.expressionator.get_value(rule.get("field_name"))
                if valid_values and field_value not in valid_values or not field_value:
                    return True
        return is_disabled

    def trigger_validator(self, validator, value):
        """_summary_

        Args:
            validator (_type_): _description_
            value (_type_): _description_

        Raises:
            ValidationError: _description_
        """
        is_valid = self.expressionator.validator(validation_data=validator, value=value)
        if not is_valid:
            raise ValidationError(detail=validator.get("default"))

        return value

    def validate_empty_value(self, value):
        """_summary_

        Args:
            value (_type_): _description_
        """

        if value in EMPTY_VALUES:
            return self.default

        return value

    def run_validation(self, value=empty):
        """_summary_

        Args:
            value (_type_, optional): _description_. Defaults to empty.

        Returns:
            _type_: _description_
        """

        if self.is_disabled_by_rules(
            enabled_rules=self.extra_kwargs.get("enabled_rules"), value=value
        ):
            return self.validate_empty_value(value=value)

        if (validator := self.extra_kwargs.get("validator")) and not (
            self.root.is_update
            and self.extra_kwargs.get("field_name") == self.root.pk_field
        ):
            self.trigger_validator(
                validator=validator,
                value=value,
            )

        return super().run_validation(self.validate_empty_value(value))

    def get_value(self, data):
        """_summary_

        Args:
            data (_type_): _description_

        Returns:
            _type_: _description_
        """
        return self.validate_empty_value(super().get_value(data))

    def to_representation(self, data):
        """_summary_

        Args:
            data (_type_): _description_

        Returns:
            _type_: _description_
        """
        return super().to_representation(self.validate_empty_value(data))


class CharField(BasicFieldSerializer, serializers.CharField):
    """
    A serializer field for handling character (string) data.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = (
        "allow_blank",
        "trim_whitespace",
        "max_length",
        "min_length",
    )
    default: Any = ""

    def __init__(self, *args, required=None, **kwargs):
        super(CharField, self).__init__(
            *args, required=required, allow_blank=not required, **kwargs
        )


class EmailField(BasicFieldSerializer, serializers.EmailField):
    """
    A serializer field for handling email addresses.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = (
        "allow_blank",
        "trim_whitespace",
        "max_length",
        "min_length",
    )
    default: Any = ""


class URLField(BasicFieldSerializer, serializers.URLField):
    """
    A serializer field for handling URLs.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = (
        "allow_blank",
        "trim_whitespace",
        "max_length",
        "min_length",
    )
    default: Any = ""


class UUIDField(BasicFieldSerializer, serializers.UUIDField):
    """
    A serializer field for handling UUIDs.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = (
        "allow_blank",
        "trim_whitespace",
        "max_length",
        "min_length",
        "format",
        "hex_verbose",
    )
    default: Any = ""


class IntegerField(BasicFieldSerializer, serializers.IntegerField):
    """
    A serializer field for handling integers.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = ("max_value", "min_value")
    default: Any = 0


class FloatField(BasicFieldSerializer, serializers.FloatField):
    """
    A serializer field for handling floats.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = ("max_value", "min_value")
    default: Any = 0.0


class DecimalField(BasicFieldSerializer, serializers.DecimalField):
    """
    A serializer field for handling decimals.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = (
        "max_digits",
        "decimal_places",
        "coerce_to_string",
        "max_value",
        "min_value",
        "normalize_output",
        "localize",
        "rounding",
    )
    default: Any = 0.0


class BooleanField(BasicFieldSerializer, serializers.BooleanField):
    """
    A serializer field for handling booleans.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = ("allow_null",)
    default: Any = False


class DateField(BasicFieldSerializer, serializers.DateField):
    """
    A serializer field for handling dates.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = ("format", "input_formats")

    def __init__(self, *args, input_formats=None, **kwargs):
        input_formats = input_formats or ["iso-8601", "%Y-%m-%dT%H:%M:%S.%fZ"]
        self.initial = self.initial if self.initial is not None else date.today()
        super().__init__(*args, input_formats=input_formats, **kwargs)


class DateTimeField(BasicFieldSerializer, serializers.DateTimeField):
    """
    A serializer field for handling date-times.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = ("format", "input_formats", "default_timezone")


class DateRangeField(BasicFieldSerializer, serializers.DateTimeField):
    """
    A custom serializer field to handle a range of dates as a list with two items.
    Expects input in the format: ["YYYY-MM-DD", "YYYY-MM-DD"]
    Outputs a list of Python date objects: [start_date, end_date]
    """

    extra_args: Tuple[str, ...] = ("format", "input_formats", "default_timezone")

    def to_internal_value(self, data):
        """_summary_

        Args:
            data (_type_): _description_

        Raises:
            serializers.ValidationError: _description_

        Returns:
            _type_: _description_
        """
        if not isinstance(data, list) or len(data) != 2:
            raise serializers.ValidationError(
                detail="Date range must be a list with two dates: [start_date, end_date]."
            )

        start_date, end_date = data
        start_date = super().to_internal_value(start_date)
        end_date = super().to_internal_value(end_date)
        return [start_date, end_date]

    def to_representation(self, value):
        """_summary_

        Args:
            value (_type_): _description_

        Raises:
            TypeError: _description_

        Returns:
            _type_: _description_
        """
        if (
            not isinstance(value, list)
            or len(value) != 2
            or not all(isinstance(d, date) for d in value)
        ):
            raise TypeError("Value must be a list with two date objects.")

        start_date, end_date = value
        start_date = super().to_representation(start_date)
        end_date = super().to_representation(end_date)
        return [start_date, end_date]


class TimeField(BasicFieldSerializer, serializers.TimeField):
    """
    A serializer field for handling time values.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = ("format", "input_formats")


class DurationField(BasicFieldSerializer, serializers.DurationField):
    """
    A serializer field for handling duration values.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args = ("max_value", "min_value")


class ListField(BasicFieldSerializer, serializers.ListField):
    """
    A serializer field for handling lists.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = ("child", "allow_empty", "max_length", "min_length")
    default: Any = []


class DictField(BasicFieldSerializer, serializers.DictField):
    """
    A serializer field for handling dictionaries.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = ("child", "allow_empty")
    default: Any = {}


class ChoiceField(BasicFieldSerializer, serializers.ChoiceField):
    """
    A serializer field for handling choices.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = (
        "choices",
        "html_cutoff",
        "html_cutoff_text",
        "allow_blank",
    )

    def __init__(self, choices=None, **kwargs):
        choices = choices or ()
        super().__init__(choices=choices, **kwargs)


class MultipleChoiceField(BasicFieldSerializer, serializers.MultipleChoiceField):
    """
    A serializer field for handling ultiple choices.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args = (
        "choices",
        "html_cutoff",
        "html_cutoff_text",
        "allow_blank",
        "allow_empty",
    )
    default = []

    def __init__(self, choices=None, **kwargs):
        choices = choices or ()
        super().__init__(choices=choices, **kwargs)

    def to_internal_value(self, data):
        return list(super().to_internal_value(data))


class SlugField(BasicFieldSerializer, serializers.SlugField):
    """
    A serializer field for handling slugs.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = ("allow_unicode",)
    default: Any = ""


class FileField(BasicFieldSerializer, serializers.FileField):
    """
    A serializer field for handling file uploads.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = ("allow_empty_file", "use_url", "max_length")


class ImageField(BasicFieldSerializer, serializers.ImageField):
    """
    A serializer field for handling image uploads.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = (
        "allow_empty_file",
        "use_url",
        "max_length",
        "_DjangoImageField",
    )


class SerializerMethodField(BasicFieldSerializer, serializers.SerializerMethodField):
    """
    A serializer field for handling method-based fields.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = ("method_name", "source", "read_only")

    def to_internal_value(self, data):
        return data


class HyperlinkedIdentityField(
    BasicFieldSerializer, serializers.HyperlinkedIdentityField
):
    """
    A serializer field for handling hyperlinked identity.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = (
        "view_name",
        "lookup_field",
        "lookup_url_kwarg",
        "format",
        "read_only",
        "source",
    )


class HyperlinkedRelatedField(
    BasicFieldSerializer, serializers.HyperlinkedRelatedField
):
    """
    A serializer field for handling hyperlinked relations.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = (
        "view_name",
        "lookup_field",
        "lookup_url_kwarg",
        "format",
    )


class PrimaryKeyRelatedField(BasicFieldSerializer, serializers.PrimaryKeyRelatedField):
    """
    A serializer field for handling primary key relations.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = (
        "pk_field",
        "queryset",
        "many",
        "read_only",
        "html_cutoff",
        "html_cutoff_text",
    )


class RelatedField(BasicFieldSerializer, serializers.RelatedField):
    """
    A serializer field for handling related objects.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = (
        "queryset",
        "many",
        "read_only",
        "html_cutoff",
        "html_cutoff_text",
    )

    def to_internal_value(self, data):
        """_summary_

        Args:
            data (_type_): _description_

        Returns:
            _type_: _description_
        """
        return data

    def to_representation(self, value):
        """_summary_

        Args:
            value (_type_): _description_

        Returns:
            _type_: _description_
        """
        return value


class JSONField(BasicFieldSerializer, serializers.JSONField):
    """
    A serializer field for handling JSON objects.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    extra_args: Tuple[str, ...] = ("binary", "encoder", "decoder")
    default: Any = {}


class ReadOnlyField(BasicFieldSerializer, serializers.ReadOnlyField):
    """
    A serializer field for read-only data.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    def to_internal_value(self, data):
        """_summary_

        Args:
            data (_type_): _description_

        Returns:
            _type_: _description_
        """
        return data


class HiddenField(BasicFieldSerializer, serializers.HiddenField):
    """
    A serializer field for hidden data.

    Attributes:
        extra_args (Tuple[str, ...]): Extra arguments for the field configuration.
        default (Any): Default value for an empty field.
    """

    def to_representation(self, value):
        """_summary_

        Args:
            value (_type_): _description_

        Returns:
            _type_: _description_
        """
        return value


class SectionFieldException(Exception):
    """_summary_

    Args:
        Exception (_type_): _description_
    """

    def __init__(self, *args, field_path=None, error=None, **kwargs) -> None:
        self.field_path = field_path
        self.error = error
        super().__init__(*args, **kwargs)


class BaseSectionField(serializers.JSONField):
    """
    _summary_

    Args:
        JSONField (_type_): _description_
    """

    extra_args = ("binary", "encoder", "decoder", "section")
    extra_kwargs = None
    section = {}
    fields = {}

    def bind(self, field_name, parent):
        """_summary_

        Args:
            field_name (_type_): _description_
            parent (_type_): _description_
        """

        super().bind(field_name, parent)

        from swifty.serializers.manager import create_field_kwargs

        self.section = self.extra_kwargs and self.extra_kwargs.get("section", {})
        self.fields = {}
        for field_props in self.section.get("fields", []):
            field_name = field_props["field_name"]
            field_base, field_kwargs = create_field_kwargs(field_props)
            field_serializer = field_base(**field_kwargs)
            field_serializer.bind(parent=self, field_name=field_name)
            self.fields[field_name] = field_serializer

    def to_internal_value(self, data):
        """_summary_

        Args:
            data (_type_): _description_

        Raises:
            ValidationError: _description_

        Returns:
            _type_: _description_
        """
        # Recursively validate nested fields
        errors = {}
        if isinstance(data, dict) and isinstance(self.fields, dict):
            for key, value in data.items():
                try:
                    data[key] = self.fields[key].run_validation(value)
                except (ValidationError, DjangoValidationError) as error:
                    errors.update(
                        {key: ValidationError(detail=validation_error_detail(error))}
                    )
                except KeyError:
                    pass

        if errors:
            raise ValidationError(detail=validation_error_detail(errors))
        return json.loads(
            super().to_internal_value(json.dumps(data, default=json_handler))
        )

    def to_representation(self, data):
        if isinstance(data, Base):
            data = json.loads(json.dumps(data, default=json_handler))

        if isinstance(data, dict) and isinstance(self.fields, dict):
            for key, value in data.items():
                if key not in self.fields:
                    continue
                data[key] = self.fields[key].to_representation(value)

        return super().to_representation(data)


class SectionField(BasicFieldSerializer, BaseSectionField):
    """
    _summary_

    Args:
        JSONField (_type_): _description_
    """

    default = {}


class SectionManyField(BasicFieldSerializer, BaseSectionField):
    """
    _summary_

    Args:
        JSONField (_type_): _description_
    """

    default = []

    def to_internal_value(self, data_list):
        """_summary_

        Args:
            data_list (_type_): _description_

        Raises:
            ValidationError: _description_

        Returns:
            _type_: _description_
        """
        # Recursively validate nested fields
        cleaned_data = []
        errors = {}
        for data in data_list:
            if isinstance(data, dict):
                try:
                    cleaned_data.append(super().to_internal_value(data))
                except (ValidationError, DjangoValidationError) as error:
                    errors.update(
                        {
                            data_list.index(data): ValidationError(
                                detail=validation_error_detail(error)
                            )
                        }
                    )

        if errors:
            raise ValidationError(detail=validation_error_detail(errors))
        return cleaned_data

    def to_representation(self, data_list):
        """_summary_

        Args:
            data_list (_type_): _description_

        Returns:
            _type_: _description_
        """
        return [
            super(SectionManyField, self).to_representation(data) for data in data_list
        ]


FIELD_TYPE_MAP: Dict[str, BasicFieldSerializer] = {
    field_serializer.__name__: field_serializer
    for field_serializer in BasicFieldSerializer.__subclasses__()
}



"""Swifty logging module"""

from django.conf import settings

if not getattr(settings, "LOGGING", None):
    from swifty.logging.config import get_logging_config

    logging_config = getattr(settings, "SWIFTY_LOGGING_CONFIG", None) or {}
    setattr(settings, "LOGGING", get_logging_config(**logging_config))



"""Decorators"""

from typing import Callable
from functools import wraps


def once_trigger(func: Callable) -> Callable:
    """Decorator to ensure a method is only called once per instance."""

    @wraps(func)
    def wrapper(self):
        saved_name = f"_{self.__class__.__name__}_{func.__name__}"
        if not hasattr(self, saved_name):
            setattr(self, saved_name, func(self))
        return getattr(self, saved_name)

    return wrapper



"""Swifty module."""

from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as f:
    requirements = [
        line.strip() for line in f if line.strip() and not line.startswith("#")
    ]

with open("VERSION", encoding="utf-8") as f:
    version = f.read().strip()


setup(
    name="django-swifty",
    version=version,
    author="Fudu",
    author_email="hphuc3005@gmail.com",
    description="Swifty Utils for Django Application",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/sampleproject",
    packages=find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Framework :: Django",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)
